#!/usr/bin/perl
use Carp;
use lib ('..');
use flo::Standard;
use lib::Exception;
use Data::Dumper;
use MKDoc;
use MKDoc::Util::Text2HTML;
use strict;
use warnings;
use 5.6.1;

=cut

BEGIN {
    $SIG{'__WARN__'} = sub {
	warn $_[0] unless ($_[0] =~ /utf8.*\.\.\/lib\/sql\/Query\.pm/);
    };
}

=cut

die "usage: $0 <hours> [\$domain]" if (@ARGV == 0);

our $INDEXABLE_RECENTNESS = int (shift (@ARGV) * 3600);
our $WEBSITE_TO_REINDEX   = shift;


try {
    main();
}
catch {
    print join "\n", lib::sql::Table->query_stack;
    my $error = (ref $@) ? (Dumper ($@)) : $@;
    die $error;
};


sub main
{
    $ENV{PATH_INFO} = '/';
    my $VAR1 = undef;
    open FP, "<../conf/sites.conf" or die "Cannot read-open sites.conf";
    my $data = join '', <FP>;
    close FP;
    $VAR1 = eval "$data" or die "Cannot eval data from sites.conf";
    
    my %websites = %{$VAR1};
    
    if (defined $WEBSITE_TO_REINDEX)
    {
	my $base_dir = $websites{$WEBSITE_TO_REINDEX};
	defined $base_dir || die "$WEBSITE_TO_REINDEX does not seem to exist";
	$ENV{SERVER_NAME} = $WEBSITE_TO_REINDEX;
	print "$base_dir\n";
        + MKDoc::init ('cgi.mkdoc');
	index_current();
    }
    else
    {
	my %setisbew = reverse %websites;
	foreach my $base_dir (keys %setisbew)
	{
	    my $website_domain = $setisbew{$base_dir};
	    $ENV{SERVER_NAME} = $website_domain;
	    print "$base_dir\n";
	    + MKDoc::init ('cgi.mkdoc');
	    index_current();
	}
    }
}


sub index_current
{
    index_remove();
    index_update();
    index_insert();
}


##
# index_remove;
# -------------
#   Remove keywords indexing documents that don't exist
##
sub index_remove
{
    my $document = table ('Document');
    my $index = table ('Document_Index');
    
    # we have to find all ids which exist in index
    # and which don't exist in $document anymore.
    my @document_ids = map { $_->{ID} => 1 } $document->select ( cols => [ qw /ID/ ] )->fetch_all;
    my %index_ids = map { $_->{Record_ID} => 1 } $index->select ( cols => [ qw /Record_ID/ ] )->fetch_all;
    
    foreach my $id (@document_ids)
    {
	delete $index_ids{$id};
    }
    
    foreach my $id (keys %index_ids)
    {
	print "\t[delete] $id\n";
	$index->delete ( Record_ID => $id );
    }
}


##
# index_update;
# -------------
#   Indexes documents which were modified recently
##
sub index_update
{
    my $document = table ('Document');
    my $index = table ('Document_Index');
    my $date = reindex_date();
    
    my $condition = new lib::sql::Condition;
    $condition->add ( 'Date_Last_Modified', '>', $date );
    my $query = $document->search ( $condition );
    
    while (my $doc = $query->next)
    {
	print "\t[update] $doc->{Full_Path}\n";
	$index->delete ( Record_ID => $doc->{ID} );
	$document->lang ($doc->{Lang});	
	
	foreach my $col (sort keys %{$doc})
	{
	    next unless (defined $document->{weight}->{$col});
	    
	    my %keywords = map { $_ => 1 } _data_split ($document, $doc->{$col});
	    my @keywords = keys %keywords;
	    
	    foreach my $keyword (@keywords)
	    {
		$index->insert ( Record_ID => $doc->{ID},
				 Column_Name => $col,
				 Keyword => $keyword );
	    }
	}
    }
}


##
# index_insert;
# -------------
#   Adds keywords for documents which are being added
##
sub index_insert
{
    my $document = table ('Document');
    my $index = table ('Document_Index');
    
    # we have to find all ids which exist in @document
    # and which have no reference in index
    my @document_ids = map { $_->{ID} } $document->select ( cols => [ qw /ID/ ] )->fetch_all;
    my %index_ids = map { $_->{Record_ID} => 1 } $index->select ( cols => [ qw /Record_ID/ ] )->fetch_all;
    
    foreach my $id (@document_ids)
    {
	(defined $index_ids{$id}) and next;
	
	my $doc = $document->get ($id) or
	    (Carp::cluck "Cannot get document with ID: $id" and next);
	
	$document->lang ($doc->{Lang});	
	print "\t[insert] $doc->{Full_Path}\n";
	
	foreach my $col (sort keys %{$doc})
	{
	    next unless (defined $document->{weight}->{$col});
	    
	    my %keywords = map { $_ => 1 } _data_split ($document, $doc->{$col});
	    my @keywords = keys %keywords;
	    
	    foreach my $keyword (@keywords)
	    {
		$index->insert ( Record_ID => $id,
				 Column_Name => $col,
				 Keyword => $keyword );
	    }
	}
    }
}


##
# reindex_date;
# -------------
#   Returns the reindex date given $INDEXABLE_RECENTNESS
##
sub reindex_date
{
    # prefill date fields with current date
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime (time - $INDEXABLE_RECENTNESS);
    $mon++;
    $year += 1900;
    
    if (length ($mon)  == 1) { $mon  = "0" . $mon  }
    if (length ($mday) == 1) { $mday = "0" . $mday }
    
    return "$year-$mon-$mday $hour:$min:$sec";	
}


##
# _data_split ($table, $value);
# -----------------------------
#   Splits the data into keywords, eventually depending on
#   the language which is being used
#
#   @param - $value : string that needs to be splitted
#   @returns - An array of splitted keywords
##
sub _data_split
{
    my $table  = shift;
    my $value = shift;
    return () unless (defined $value);
    
    # convert all the HTML to text first.
    $value = html2text ($value);
    
    # make $value ASCII
    my $lang = $table->lang;
    my $lang_util_object = MKDoc::Lang->create ($lang);
    
    return map { (length ($_) > 1) ? $_ : () } $lang_util_object->ascii_keywords ($value);
}


1;


