package HTML_Page;
use URI;
use LWP::UserAgent;
use HTTP::Request;
use HTTP::Response;
use HTTP::Headers;
use HTML::LinkExtor;
use Date::Manip;
use HTML::TreeBuilder;
use strict;
use warnings;
use Unicode::String;


# returns a new HTML_Page object
sub new
{
    my $class = shift;
    $class = ref $class || $class;
    return bless { @_ }, $class;
}


# returns the canonical URI of this HTML_Page object
sub uri
{
    my $self = shift;
    if (@_) 
    {
	my $uri = URI->new (shift)->canonical;
	$self->{uri} = shift;
    }
    
    return $self->{uri};
}


sub full_path
{
    my $self = shift;
    URI->new ($self->uri)->path;
}


sub name
{
    my $self = shift;
    my $path = $self->full_path;
    $path =~ s/^\///;
    $path =~ s/\/$//;
    if ($path)
    {
	if ($path =~ /\//)
	{
	    my @split = split /\//, $path;
	    return pop (@split);
	}
	else
	{
	    return $path;
	}
    }
    else
    {
	return;
    }
}


# this should be called before this HTML_Page object is
# serialized and written onto disk
sub before_save
{
    my $self = shift;
    delete $self->{fresh};
}


# returns TRUE if this HTML_Page is ok (i.e. the
# remote server serves it), FALSE otherwise
sub is_ok
{
    my $self = shift;
    $self->load;
    return $self->{code} =~ /^2/;
}


# returns this HTML_Page title
sub title
{
    my $self = shift;
    $self->load;
    $self->is_ok or return;
    my $data = $self->{content};
    my ($res) = $data =~ /\<title\>(.*?)\<\/title\>/ism;
    $res =~ s/^(\s|\r|\n)+//gsm;
    $res =~ s/(\s|\r|\n)+$//gsm;
    return $res;
}


# returns this HTML_Page language ISO code
sub lang
{
    my $self = shift;
    $self->load;
    $self->is_ok or return;
    my $data = $self->{content};
    my ($res) = $data =~ /\"DC.Lang\".*?content\=\"(.*?)\"/gsm;
    $res =~ s/^(\s|\r|\n)+//gsm;
    $res =~ s/(\s|\r|\n)+$//gsm;
    return $res || 'en';
}


# returns this HTML_Page description
sub description
{
    my $self = shift;
    $self->load;
    $self->is_ok or return;
    my $data = $self->{content};
    my ($res) = $data =~ /\<\s*meta\s+name\s*=\s*\"\s*description\s*\"\s+content\s*=\s*\"(.*?)\"/ism;
    return '' unless (defined $res);
    $res =~ s/^(\s|\r|\n)+//gsm;
    $res =~ s/(\s|\r|\n)+$//gsm;
    return $res;
}


# returns this HTML_Page keywords
sub keywords
{
    my $self = shift;
    $self->load;
    $self->is_ok or return;
    my $data = $self->{content};
    my ($res) = $data =~ /\<\s*meta\s+name\s*=\s*\"\s*keywords\s*\"\s+content\s*=\s*\"(.*?)\"/ism;
    return '' unless (defined $res);
    $res =~ s/^(\s|\r|\n)+//gsm;
    $res =~ s/(\s|\r|\n)+$//gsm;
    return $res;
}


# returns this HTML_Page's creation_date
sub date_created
{
    my $self = shift;
    $self->load;
    $self->is_ok or return;
    my $data = $self->{content};
    my ($res) = $data =~ /\"DC.Date.created\".*?content\=\"(.*?)\"/gsm;
    $res ||= '';
    $res =~ s/^(\s|\r|\n)+//gsm;
    $res =~ s/(\s|\r|\n)+$//gsm;
    return $res || '2002-01-01';
}


# returns this HTML_Page's last_modified_date
sub modified_date
{
    my $self = shift;
    $self->load;
    $self->is_ok or return;
    my $data = $self->{content};
    my ($res) = $data =~ /\"DC.Date.modified\".*?content\=\"(.*?)\"/gsm;
    $res ||= '';
    $res =~ s/^(\s|\r|\n)+//gsm;
    $res =~ s/(\s|\r|\n)+$//gsm;
    return $res || '2002-01-01';
}


# returns this HTML_Page's creation_date, but as MySQL
sub date_created_mysql
{
    my $self = shift;
    my $date_created = &ParseDate ($self->date_created);
    return &UnixDate ($date_created, "%Y-%m-%d");
}


# returns this HTML_Page's creation_date, but as MySQL
sub date_modified_mysql
{
    my $self = shift;
    my $date_created = &ParseDate ($self->date_created);
    return &UnixDate ($date_created, "%Y-%m-%d");
}


# returns this HTML_Page body
sub body
{
    my $self = shift;
    $self->load;
    $self->is_ok or return;
    my $data = $self->{content};
    my ($res) = $data =~ /\<\s*body.*?>(.*?)<\/body>/ism;
    return '' unless (defined $res);
    $res =~ s/^(\s|\r|\n)+//gsm;
    $res =~ s/(\s|\r|\n)+$//gsm;
    return $res;
}


# returns this HTML_Page links as an array
sub links
{
    my $self = shift;
    $self->load;
    $self->is_ok or return;
    
    my %res  = ();
    my $data = $self->{content};
    my $tree = new HTML::TreeBuilder;
    $tree->parse ($data);
    $tree->elementify;
    my @nodes = ($tree);
    while (@nodes)
    {
	my $node = shift (@nodes);
	next unless (ref $node);
	push @nodes, $node->content_list();
	if ($node->tag eq 'a')
	{
	    my $href  = $node->attr ('href');
	    my $lang  = $node->attr ('lang') || $node->attr ('xml:lang');
	    
	    my $base_uri = new URI ($self->uri);
	    my $href_uri = new URI ($href);
	    $href_uri = $href_uri->abs ($base_uri);
	    $href = $href_uri->canonical;
	    next unless (defined $href_uri->scheme);
	    next unless ($href_uri->scheme =~ /^https?/);
	    $res{$href} = 1;
	}
    }
    return sort (keys %res);
}


# returns this HTML_Page INTERNAL links as an array
sub internal_links
{
    my $self = shift;
    my @links = $self->links;
    my @res = ();
    my $self_URI = new URI ($self->uri);
    foreach my $link (@links)
    {
	my $link_URI = new URI ($link);
	push @res, $link if ($self_URI->host eq $link_URI->host);
    }
    return @res;
}


# returns this HTML_Page EXTERNAL links as an array
sub external_links
{
    my $self = shift;
    my @links = $self->links;
    my @res = ();
    my $self_URI = new URI ($self->uri);
    foreach my $link (@links)
    {
	my $link_URI = new URI ($link);
	push @res, $link if ($self_URI->host ne $link_URI->host);
    }
    return @res;
}


# loads this HTML_Page into memory
sub load
{
    my $self = shift;
    return if ($self->{fresh});
    
    if ($self->{content})
    {
	my $ua = LWP::UserAgent->new;
	my $request = HTTP::Request->new ( HEAD => $self->uri );
	my $response = $ua->request ($request);
	
	my $code = $response->code;
	my $headers = $response->headers;
	
	my $date = $headers->{'Last-Modified'};
	$date = &ParseDate ($date) if (defined $date and $date);
	
	my $etag = $headers->{'ETag'};
	my $md5 = $headers->{'Content-MD5'};
	
	if (defined $self->{md5} and defined $md5 and $self->{md5} eq $md5)
	{
	    $self->{fresh} = 1;
	    return;
	}
	
	if (defined $self->{etag} and defined $etag and $self->{etag} eq $etag)
	{
	    $self->{fresh} = 1;
	    return;
	}
	
	if (defined $self->{date} and defined $date and $self->{date} eq $date)
	{
	    $self->{fresh} = 1;
	    return;
	}
    }
    
    my $ua = LWP::UserAgent->new;
    my $request = HTTP::Request->new ( GET => $self->uri );
    my $response = $ua->request ($request);
    
    my $code = $response->code;
    my $content = $response->content;
    my $headers = $response->headers;
    
    my $date = $headers->{'Last-Modified'};
    my $etag = $headers->{'ETag'};
    my $md5  = $headers->{'Content-MD5'};
    
    $self->{code} = $code;
    $self->{date} = $date;
    $self->{etag} = $etag;
    $self->{md5} = $md5;
    $self->{content} = $content;
    
    $self->{uri} = $response->request->uri;
    $self->{fresh} = 1;
}


1;
